#ifndef GENERAL_H
#define GENERAL_H

#pragma warning(disable : 4996)

#include <stdlib.h>
#include <stdio.h>
#include <fstream>
#include <iostream>
#include <string>
#include <cmath>
#include <time.h>

// chemin et nom du fichier comportant les données de l'instance
#define NOMFICHIERINSTANCE "instances/plus de 200 villes/VRP_DLP_18.txt"


// pas oublier d'arrondir ce chiffre à la dizaine au-dessus de la taille de l'instance demandée ci-dessus, par exemple 110 pour une instance de 104 villes.
#define NBRMAXNOEUDS 260

// pour l'instance "instances/en-dessous de 100 villes/VRP_DLP_01.txt" (91 villes), ces paramètres sont un bon compromis d'après mes expériences. De très bons résultats pour environ 1 heure d'exécution :
//          GROSSEOPTI_TSP_INITIALE = true
//				    NBRPREMIERCYCLE = 30
//                 NBRDEUXIEMECYCLE = 3
// NBRSOLUTIONSVRPPOURDEUXIEMECYCLE = 5
//     NBRDERUNSREGRESSIONRECURSIVE = 10
//    PROFONDEURREGRESSIONRECURSIVE = 3
//   NBRBRANCHESREGRESSIONRECURSIVE = 7

// Or la complexité (donc durée d'exécution) croît de façon exponentielle avec la taille de l'instance.
// Par exemple, les paramètres mêmes ci-dessus trouvent presque instantanément plusieurs fois le minimum global pour l'instance "instances/en-dessous de 100 villes/VRP_DLP_75.txt" (19 villes) alors que ils n'ont pas du tout été suffisants pour trouver le minimum global de l'instance 01 (91 villes) en 1 heure d'exécution.
// Il faut donc jouer avec les paramètres ci-dessous en fonction de la taille de l'instance pour ne pas que le temps d'exécution soit trop long mais tout de même garder une certaine qualité des résultats.
// Il y a aussi une autre technique qui consiste en désactiver la fonction d'optimisation des solutions TSP par RL_deplacementSommetsTSP dans la fonction productionDeToursGeantsPourSPLIT(). Le bout de code en question se trouve dans outils.cpp à la ligne 1282 environ. 
// aussi, il ne faut pas hésiter à exécuter ce programme avec la commande nohup si on veut fermer le terminal ou faire autre chose en attendant sur sa machine. mode "tâche de fond avec l'esperluette &". Par exemple exécuter la commande "nohup ./projetVRP &"
// exécuter avec nohup signifie que la sortie standard (ce qui s'écrit dans le terminal) est redirigé vers un fichier "nohup.out" qui se crée automatiquement dans le répertoire courant et qui est consultable dès que l'exécution est terminée. 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Détermine si oui ou non le programme doit optimiser au maximum les solutions TSP initiales avant SPLIT (après phase de construction) car cette fonctionnalité est très gourmande, pas très utile et pas paramètrable.
// Cependant cette phase d'optimisation a de fortes chances de trouver le meilleur tour géant possible pour des instances de moins de 100 villes, ce qui n'est pas chose facile. 
#define GROSSEOPTI_TSP_INITIALE false

// à peu près linéaire, ce ne sont pas les paramètres les plus sensibles
#define NBRPREMIERCYCLE 3
#define NBRDEUXIEMECYCLE 2
#define NBRSOLUTIONSVRPPOURDEUXIEMECYCLE 5
#define NBRDERUNSREGRESSIONRECURSIVE 5

// attention ! la complexité de cette portion de code dépend notamment de O(A^0 + A^1 + A^2 + … + A^B) avec A = NBRBRANCHESREGRESSIONRECURSIVE et B = PROFONDEURREGRESSIONRECURSIVE. Cela peut donc monter en complexité très très vite
#define PROFONDEURREGRESSIONRECURSIVE 3
#define NBRBRANCHESREGRESSIONRECURSIVE 4

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#define PROBAPRINTPATIENCE powf(NBRBRANCHESREGRESSIONRECURSIVE, PROFONDEURREGRESSIONRECURSIVE - 0.4)
#define NBRALGOSPERTURBATIONS 3
typedef enum 
{
	PERMUTER,
	INVERSER,
	DEPLACER
}typePerturbation_e;


#define SOMMETDEPOT 0

#define NBRPLUSPROCHESVOISINS 5
#define PROBAPREMIERCHOIX 80
#define SUR10 100

#define TAILLEPREFIXESOLUTION 11
#define TAILLEPLUSGRANDNOMSOLUTION 56

#define NBRALGOSCONSTRUCTION_TSP 6
#define NBRALGOSCONSTRUCTION_VRP 1
#define TAILLEMAXCONTENANTSOLUTION_TSP 2 * NBRALGOSCONSTRUCTION_TSP // 2 * Nombre d'algos de construction de tour géant
#define TAILLEMAXCONTENANTSOLUTION_VRP TAILLEMAXCONTENANTSOLUTION_TSP + NBRALGOSCONSTRUCTION_VRP // NBRCONSTRUCTIONSTSP + Nombre d'algos de construction de VRP

#define CONSTANTEELLIPSE 6
#define CONSTANTERAYONCERCLE 0.5
#define PROPORTIONCAPACITYVEHICLE 0.5

typedef struct typeInstance
{
	int nbrNoeudsInstance;
	int capacityVehicle;
	int ** distances;
	int * quantityCollectable;
}instance_t;

typedef struct typeSolutionTSP
{
	std::string nomSolutionTSP;
	int coutTotalTSP;
	int quantityTotaleTSP;
	int nbrNoeudsTSP;
	int * etapesTSP;
}solutionTSP_t;

typedef struct typeSolutionVRP
{
	std::string nomSolutionVRP;
	int nbrTourneesVRP;
	int coutTotalVRP;
	int quantityTotaleVRP;
	solutionTSP_t * tableauTSP;
}solutionVRP_t;

typedef struct typeContenantTableauSolutionsTSP
{
	std::string nomTableauSolutionsTSP;
	int nbrSolutionsTSP;
	solutionTSP_t * tableauSolutionsTSP;
}contenantTableauSolutionsTSP_t;

typedef struct typeContenantTableauSolutionsVRP
{
	std::string nomTableauSolutionsVRP;
	int nbrSolutionsVRP;
	solutionVRP_t * tableauSolutionsVRP;
}contenantTableauSolutionsVRP_t;

#endif