/****************** TSP, VRP et Recherche Opérationnelle ****************
        --> TP3 d'Outils d'Aide à la décision | S1 ZZ2 F2
------------------------------------------------------------------------
	Code écrit par Anaël Courjaud en Décembre 2022 - Janvier 2023
		Dernière modification le 18/01/2023 par Anaël Courjaud
------------------------------------------------------------------------*/

#include "general.h"
#include "outils.h"
#include "allocDynamiqueEtAffichage.h"

int main()
{
    time_t begin = time( NULL ); // On enregistre le début du programme.

	srand((unsigned int)time(NULL)); // Initialisation pour la fonction rand()


	std::cout << "\n\n\n****************** TSP, VRP et Recherche Opérationnelle ****************\n";
	std::cout << "        --> TP3 d'Outils d'Aide à la décision | S1 ZZ2 F2\n";
	std::cout << "------------------------------------------------------------------------\n";
	std::cout << "	Code écrit par Anaël Courjaud en Décembre 2022 - Janvier 2023\n";
	std::cout << "		Dernière modification le 18/01/2023 par Anaël Courjaud\n";
	std::cout << "------------------------------------------------------------------------\n";
	std::cout << "\n\n\n";


	instance_t instance;
	fonctionMalloc(instance);
	instanciation(NOMFICHIERINSTANCE, instance);


	contenantTableauSolutionsTSP_t contenantTableauMeilleuresSolutionsTSPpourSPLIT;
	fonctionMalloc(contenantTableauMeilleuresSolutionsTSPpourSPLIT, TAILLEMAXCONTENANTSOLUTION_TSP);

	contenantTableauSolutionsVRP_t contenantTableauSolutionsVRPpour1erCycle;
	fonctionMalloc(contenantTableauSolutionsVRPpour1erCycle, TAILLEMAXCONTENANTSOLUTION_VRP);

	std::cout << "\nRun production de tours géants pour SPLIT\n";
	productionDeToursGeantsPourSPLIT(contenantTableauMeilleuresSolutionsTSPpourSPLIT, instance);
	std::cout << "\nRun production de VRP pour 1er cycle d'optimisation (grâce à SPLIT et une construction directe de solution VRP)\n\n";
	productionDeVRPpour1erCycleOpti(contenantTableauMeilleuresSolutionsTSPpourSPLIT, contenantTableauSolutionsVRPpour1erCycle, instance);

	contenantTableauSolutionsVRP_t contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle;
	fonctionMalloc(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle, NBRPREMIERCYCLE * contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP);
	contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nomTableauSolutionsVRP = "contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle";
	contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP = 0;


	std::cout << "Entrée dans la boucle qui lance NBRPREMIERCYCLE = " << NBRPREMIERCYCLE << " fois le 1er cycle d'optimisation sur les solutions VRP fraîchement splitées (ou directement construites)\n";
	for (int i = 0; i < NBRPREMIERCYCLE; i++)
	{
		std::cout << "	Run numéro " << i + 1 << " du 1er cycle d'optimisation, plus que " << NBRPREMIERCYCLE - i - 1 << "\n";

		contenantTableauSolutionsVRP_t meilleuresSolutionsVRPtrouvees;
		fonctionMalloc(meilleuresSolutionsVRPtrouvees, contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP);
		meilleuresSolutionsVRPtrouvees.nomTableauSolutionsVRP = "MeilleuresSolutionsVRPtrouvees1erCycle_Run" + std::to_string(i);
		meilleuresSolutionsVRPtrouvees.nbrSolutionsVRP = 0;

		procedure1erCycleOptimisation(instance, contenantTableauSolutionsVRPpour1erCycle, meilleuresSolutionsVRPtrouvees);

		for (int j = 0; j < contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP; j++)
		{
			copierSolutionVRP(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.tableauSolutionsVRP[j + i * contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP], meilleuresSolutionsVRPtrouvees.tableauSolutionsVRP[j]);
			contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP++;
		}

		fonctionFree(meilleuresSolutionsVRPtrouvees, contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP);
	}

	// On sélectionne seulement les NBRSOLUTIONSVRPPOURDEUXIEMECYCLE meilleures solutions en sortie du 1er cycle d'opti pour injecter en entrée du le 2eme cycle d'opti.
	triCoutCroissantContenantTableauSolutionsVRP(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle);
	contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP = NBRSOLUTIONSVRPPOURDEUXIEMECYCLE;

	contenantTableauSolutionsVRP_t contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle;
	fonctionMalloc(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle, NBRDEUXIEMECYCLE * NBRSOLUTIONSVRPPOURDEUXIEMECYCLE);
	contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.nomTableauSolutionsVRP = "contenantTableauSauvegardeSolutionsVRPtrouveesAvecTousLes2emeCycle";
	contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.nbrSolutionsVRP = 0;

	for (int i = 0; i < NBRDEUXIEMECYCLE; i++)
	{
		std::cout << "\n\nRun numéro " << i + 1 << " du 2ème cycle d'optimisation, plus que " << NBRDEUXIEMECYCLE - i - 1 << "\n";

		contenantTableauSolutionsVRP_t meilleuresSolutionsVRPtrouvees;
		fonctionMalloc(meilleuresSolutionsVRPtrouvees, contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP);
		meilleuresSolutionsVRPtrouvees.nomTableauSolutionsVRP = "MeilleuresSolutionsVRPtrouveesAvec2emeCycle_Run" + std::to_string(i);
		meilleuresSolutionsVRPtrouvees.nbrSolutionsVRP = 0;

		// Optimisation avec recherche par régressions successives   !!TRES LONG!!   (conseil : NBRDERUNSREGRESSIONRECURSIVE <= 5; PROFONDEURREGRESSIONRECURSIVE <= 3; NBRBRANCHESREGRESSIONRECURSIVE <= 10)
		//  -----------------------------------------------------------------------------------------------------------------------------------------------------------------------
		procedure2emeCycleOptimisation(instance, contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle, meilleuresSolutionsVRPtrouvees);
		// -----------------------------------------------------------------------------------------------------------------------------------------------------------------------

		for (int j = 0; j < contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP; j++)
		{
			copierSolutionVRP(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.tableauSolutionsVRP[j + i * contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP], meilleuresSolutionsVRPtrouvees.tableauSolutionsVRP[j]);
			contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.nbrSolutionsVRP++;
		}

		std::cout << "\n";
		triCoutCroissantContenantTableauSolutionsVRP(meilleuresSolutionsVRPtrouvees);
		afficherContenantTableauSolutionsVRP(meilleuresSolutionsVRPtrouvees);

		fonctionFree(meilleuresSolutionsVRPtrouvees, contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle.nbrSolutionsVRP);
	}



	// Un peu de mise en page des résultats : 
	std::cout << "\n\nProgramme exécuté sur l'instance '" << NOMFICHIERINSTANCE << "' (" << instance.nbrNoeudsInstance << ") avec les paramètres suivants :\n\n";
	std::cout << "         GROSSEOPTI_TSP_INITIALE = "; if(GROSSEOPTI_TSP_INITIALE){std::cout << "true\n";}else{std::cout << "false\n";}
	std::cout << "                 NBRPREMIERCYCLE = " << NBRPREMIERCYCLE << "\n";
	std::cout << "                NBRDEUXIEMECYCLE = " << NBRDEUXIEMECYCLE << "\n";
	std::cout << "NBRSOLUTIONSVRPPOURDEUXIEMECYCLE = " << NBRSOLUTIONSVRPPOURDEUXIEMECYCLE << "\n";
	std::cout << "    NBRDERUNSREGRESSIONRECURSIVE = " << NBRDERUNSREGRESSIONRECURSIVE << "\n";
	std::cout << "   PROFONDEURREGRESSIONRECURSIVE = " << PROFONDEURREGRESSIONRECURSIVE << "\n";
	std::cout << "  NBRBRANCHESREGRESSIONRECURSIVE = " << NBRBRANCHESREGRESSIONRECURSIVE << "\n";
	std::cout << "\n\n";

	// On affiche en premier lieu dans l'ordre de cout croissant les constructions TSP optimisées qui ont été élaborées en début d'algo (on a enlevé les doublons et rajouté les inverses avant de rentrer tout ça dans le SPLIT)
	triCoutCroissantContenantTableauSolutionsTSP(contenantTableauMeilleuresSolutionsTSPpourSPLIT);
	afficherContenantTableauSolutionsTSP(contenantTableauMeilleuresSolutionsTSPpourSPLIT);

	// Ensuite on affiche les solutions VRP produites juste avant le premier cycle d'optimisation (sans régression)
	triCoutCroissantContenantTableauSolutionsVRP(contenantTableauSolutionsVRPpour1erCycle);
	afficherContenantTableauSolutionsVRP(contenantTableauSolutionsVRPpour1erCycle);

	// On enchaîne avec les NBRSOLUTIONSVRPPOURDEUXIEMECYCLE meilleures solution VRP issues du premier cycle d'optimisation (sans régression)
	triCoutCroissantContenantTableauSolutionsVRP(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle);
	afficherContenantTableauSolutionsVRP(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle);

	// Enfin on affiche les NBRDEUXIEMECYCLE * NBRSOLUTIONSVRPPOURDEUXIEMECYCLE solutions trouvées avec le 2ème cycle d'optimisation (avec régression)
	triCoutCroissantContenantTableauSolutionsVRP(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle);
	afficherContenantTableauSolutionsVRP(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle);

	// Et pour finir, on affiche le détail de la meilleure solutionVRP trouvée au terme des deux cycles d'optimisation!!
	std::cout << "\n\nVoici le détail de la meilleure solution VRP trouvée avec ses solutions TSP (tournées) successives :\n\n";
	afficherSolutionVRP(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.tableauSolutionsVRP[0]);
	std::cout << "\n";
	afficherTourneesDeSolutionVRP(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle.tableauSolutionsVRP[0]);


	// On oublie pas de free toutes les allocs pour éviter les fuites de mémoire
	fonctionFree(contenantTableauMeilleuresSolutionsTSPpourSPLIT, TAILLEMAXCONTENANTSOLUTION_TSP);
	fonctionFree(contenantTableauSolutionsVRPpour1erCycle, TAILLEMAXCONTENANTSOLUTION_VRP);
	fonctionFree(contenantTableauSauvegardeMeilleuresSolutionsVRPpour2emeCycle, NBRPREMIERCYCLE * contenantTableauSolutionsVRPpour1erCycle.nbrSolutionsVRP);
	fonctionFree(contenantTableauSauvegardeSolutionsVRPtrouveesAvec2emeCycle, NBRDEUXIEMECYCLE * NBRSOLUTIONSVRPPOURDEUXIEMECYCLE);
	fonctionFree(instance);


	// On note la date de fin d'exécution.
	time_t end = time( NULL);

	// Impression de la date de compilation, la date de début d'exécution, la date de fin d'exécution, ainsi que le temps pris par l'exécution.
	std::string dateDebut = ctime(&begin);
    std::string dateFin = ctime(&end);

	std::cout << "Compilation effectuée le " << __DATE__ << " à " << __TIME__ << "\n";
    std::cout << "Exécution lancée le " << dateDebut;
    std::cout << "Exécution achevée le " << dateFin;

	unsigned long secondes = (unsigned long) difftime( end, begin );

    int jours = (int)(secondes / 86400);
    int secondesRestantes = (int)(secondes % 86400);
    int heures = secondesRestantes / 3600;
    secondesRestantes %= 3600;
    int minutes = secondesRestantes / 60;
    secondesRestantes %= 60;

    std::cout << "\nLe programme a donc été exécuté en " << secondes << " secondes : " << jours << " jours, " << heures << " heures, " << minutes << " minutes et " << secondesRestantes << " secondes\n";  
        

	return 0;
}