#include "outils.h"
#include "allocDynamiqueEtAffichage.h"

void instanciation(std::string nomFichier, instance_t &instance)

{
	std::fstream file(nomFichier);

	if (file.is_open())
	{
		std::cout << "Fichier " << NOMFICHIERINSTANCE << " ouvert\n";
	}
	else
	{
		std::cout << "Echec de l'ouverture du fichier" << NOMFICHIERINSTANCE << "\n";
		exit(-1);
	}

	int nbrNoeuds = 0;
	file >> nbrNoeuds;
	instance.nbrNoeudsInstance = nbrNoeuds;

	int capacityVehicle = 0;
	file >> capacityVehicle;
	instance.capacityVehicle = capacityVehicle;

	int indiceNoeudDepot = 0;
	file >> indiceNoeudDepot;

	for (int i = 0; i <= nbrNoeuds; i++)
	{
		for (int j = 0; j <= nbrNoeuds; j++)
		{
			int distance = 0;
			file >> distance;
			instance.distances[i][j] = distance;
		}
	}

	instance.quantityCollectable[indiceNoeudDepot] = 0;
	for (int i = 0; i < nbrNoeuds; i++)
	{
		int indiceNoeud = 0;
		int quantityCollectable = 0;

		file >> indiceNoeud;
		file >> quantityCollectable;

		instance.quantityCollectable[indiceNoeud] = quantityCollectable;
	}

	file.close();
}

void constructionTSPplusProcheVoisins(instance_t instance, solutionTSP_t &solutionTSP)
{
	constructionTSPavecTousLesSommetsIndiceCroissant(instance, solutionTSP);
	solutionTSP.nomSolutionTSP = "solutionTSPplusProcheVoisin";

	for (int indiceNoeudCourant = 0; indiceNoeudCourant < solutionTSP.nbrNoeudsTSP - 1; indiceNoeudCourant++) // pour tous les noeuds existants
	{
		int indiceGoPermuter = indiceNoeudPlusProcheVoisinNotVisited(instance, solutionTSP, indiceNoeudCourant); // trouver le plus proche voisin du noeud courant

		goPermuterDansSolutionTSP(solutionTSP, indiceGoPermuter, indiceNoeudCourant + 1); // et le permuter avec le suivant du noeud courant
	}

	evaluerTSP(instance, solutionTSP, false);
}

void constructionTSPplusLointainVoisin(instance_t instance, solutionTSP_t &solutionTSP)
{
	constructionTSPavecTousLesSommetsIndiceCroissant(instance, solutionTSP);
	solutionTSP.nomSolutionTSP = "solutionTSPplusLointainVoisin";

	for (int indiceNoeudCourant = 0; indiceNoeudCourant < solutionTSP.nbrNoeudsTSP - 1; indiceNoeudCourant++) // pour tous les noeuds existants
	{
		int indiceGoPermuter = indiceNoeudPlusLointainVoisinNotVisited(instance, solutionTSP, indiceNoeudCourant); // trouver le plus lointain voisin du noeud courant

		goPermuterDansSolutionTSP(solutionTSP, indiceGoPermuter, indiceNoeudCourant + 1); // et le permuter avec le suivant du noeud courant
	}

	evaluerTSP(instance, solutionTSP, false);
}

void constructionTSPrandom(instance_t instance, solutionTSP_t &solutionTSP)
{
	solutionTSP.nomSolutionTSP = "solutionTSPrandom";

	int sommetsRestants[NBRMAXNOEUDS];
	for (int i = 0; i < instance.nbrNoeudsInstance; i++)
	{
		sommetsRestants[i] = i + 1;
	}
	int nbrSommetsRestants = instance.nbrNoeudsInstance;

	solutionTSP.etapesTSP[0] = SOMMETDEPOT;
	int indiceSommetCourant = 1;

	while (nbrSommetsRestants > 0)
	{
		int tirageIndice = rand() % nbrSommetsRestants;
		solutionTSP.etapesTSP[indiceSommetCourant] = sommetsRestants[tirageIndice];
		indiceSommetCourant++;

		for (int i = tirageIndice; i < nbrSommetsRestants - 1; i++) // décalage à gauche
		{
			if (i + 1 < NBRMAXNOEUDS)
			{
				sommetsRestants[i] = sommetsRestants[i + 1];
			}
		}
		nbrSommetsRestants--;
	}

	solutionTSP.etapesTSP[indiceSommetCourant] = SOMMETDEPOT;
	solutionTSP.nbrNoeudsTSP = indiceSommetCourant - 1;

	evaluerTSP(instance, solutionTSP, false);
}

int indiceNoeudPlusProcheVoisinNotVisited(instance_t instance, solutionTSP_t &solutionTSP, int indiceNoeudCourant)
{
	int etapeIndiceNoeudCourant = solutionTSP.etapesTSP[indiceNoeudCourant];
	int iMin = indiceNoeudCourant + 1;
	int min = instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[iMin]];
	for (int i = iMin + 1; i <= solutionTSP.nbrNoeudsTSP; i++)
	{
		if (instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[i]] < min)
		{
			iMin = i;
			min = instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[iMin]];
		}
	}

	return iMin;
}

int indiceNoeudPlusLointainVoisinNotVisited(instance_t instance, solutionTSP_t &solutionTSP, int indiceNoeudCourant)
{
	int etapeIndiceNoeudCourant = solutionTSP.etapesTSP[indiceNoeudCourant];
	int iMax = indiceNoeudCourant + 1;
	int max = instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[iMax]];
	for (int i = iMax + 1; i <= solutionTSP.nbrNoeudsTSP; i++)
	{
		if (instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[i]] > max)
		{
			iMax = i;
			max = instance.distances[etapeIndiceNoeudCourant][solutionTSP.etapesTSP[iMax]];
		}
	}

	return iMax;
}

void evaluerTSP(instance_t instance, solutionTSP_t &solutionTSP, bool affichage)
{
	solutionTSP.coutTotalTSP = 0;
	solutionTSP.quantityTotaleTSP = 0;

	for (int i = 0; i <= solutionTSP.nbrNoeudsTSP; i++)
	{
		solutionTSP.coutTotalTSP += instance.distances[solutionTSP.etapesTSP[i]][solutionTSP.etapesTSP[i + 1]];
		solutionTSP.quantityTotaleTSP += instance.quantityCollectable[solutionTSP.etapesTSP[i]];
	}

	if (affichage)
	{
		afficherSolutionTSP(solutionTSP);
	}
}

void evaluerVRP(instance_t instance, solutionVRP_t &solutionVRP, bool affichage)
{
	solutionVRP.coutTotalVRP = 0;
	solutionVRP.quantityTotaleVRP = 0;

	if (affichage)
	{
		std::cout << "!! " << solutionVRP.nomSolutionVRP << " !!\n";
		std::cout << "----------------------------------\n";
	}

	for (int i = 0; i < solutionVRP.nbrTourneesVRP; i++)
	{
		if (affichage)
		{
			evaluerTSP(instance, solutionVRP.tableauTSP[i], true);
		}
		solutionVRP.coutTotalVRP += solutionVRP.tableauTSP[i].coutTotalTSP;
		solutionVRP.quantityTotaleVRP += solutionVRP.tableauTSP[i].quantityTotaleTSP;
	}

	if (affichage)
	{
		std::cout << "----------------------------------\n";
		afficherSolutionVRP(solutionVRP);
	}
}

void triCoutCroissantContenantTableauSolutionsTSP(contenantTableauSolutionsTSP_t &contenantTableauSolutionsTSP)
{
	for (int j = 0; j < contenantTableauSolutionsTSP.nbrSolutionsTSP - 1; j++)
	{
		int iMin = j;
		int min = contenantTableauSolutionsTSP.tableauSolutionsTSP[iMin].coutTotalTSP;

		for (int i = j + 1; i < contenantTableauSolutionsTSP.nbrSolutionsTSP; i++)
		{
			if (contenantTableauSolutionsTSP.tableauSolutionsTSP[i].coutTotalTSP < min)
			{
				iMin = i;
				min = contenantTableauSolutionsTSP.tableauSolutionsTSP[iMin].coutTotalTSP;
			}
		}

		if (iMin != j)
		{
			solutionTSP_t solutionTSPtemp;
			fonctionMalloc(solutionTSPtemp);

			copierSolutionTSP(solutionTSPtemp, contenantTableauSolutionsTSP.tableauSolutionsTSP[j]);
			copierSolutionTSP(contenantTableauSolutionsTSP.tableauSolutionsTSP[j], contenantTableauSolutionsTSP.tableauSolutionsTSP[iMin]);
			copierSolutionTSP(contenantTableauSolutionsTSP.tableauSolutionsTSP[iMin], solutionTSPtemp);

			fonctionFree(solutionTSPtemp);
		}
	}
}

void triCoutCroissantContenantTableauSolutionsVRP(contenantTableauSolutionsVRP_t &contenantTableauSolutionsVRP)
{
	for (int j = 0; j < contenantTableauSolutionsVRP.nbrSolutionsVRP - 1; j++)
	{
		int iMin = j;
		int coutMin = contenantTableauSolutionsVRP.tableauSolutionsVRP[iMin].coutTotalVRP;

		for (int i = j + 1; i < contenantTableauSolutionsVRP.nbrSolutionsVRP; i++)
		{
			if (contenantTableauSolutionsVRP.tableauSolutionsVRP[i].coutTotalVRP < coutMin)
			{
				iMin = i;
				coutMin = contenantTableauSolutionsVRP.tableauSolutionsVRP[iMin].coutTotalVRP;
			}
		}

		if (iMin != j)
		{
			solutionVRP_t solutionVRPtemp;
			fonctionMalloc(solutionVRPtemp);

			copierSolutionVRP(solutionVRPtemp, contenantTableauSolutionsVRP.tableauSolutionsVRP[j]);
			copierSolutionVRP(contenantTableauSolutionsVRP.tableauSolutionsVRP[j], contenantTableauSolutionsVRP.tableauSolutionsVRP[iMin]);
			copierSolutionVRP(contenantTableauSolutionsVRP.tableauSolutionsVRP[iMin], solutionVRPtemp);

			fonctionFree(solutionVRPtemp);
		}
	}
}

void trierTourneesVRPparQuantityTotaleTSPdecroissante(solutionVRP_t &solutionVRP)
{
	for (int j = 0; j < solutionVRP.nbrTourneesVRP - 1; j++)
	{
		int iMax = j;
		int coutMax = solutionVRP.tableauTSP[iMax].quantityTotaleTSP;

		for (int i = j + 1; i < solutionVRP.nbrTourneesVRP; i++)
		{
			if (solutionVRP.tableauTSP[i].quantityTotaleTSP > coutMax)
			{
				iMax = i;
				coutMax = solutionVRP.tableauTSP[iMax].quantityTotaleTSP;
			}
		}

		solutionTSP_t solutionTSPtemp = solutionVRP.tableauTSP[j];
		fonctionMalloc(solutionTSPtemp);

		copierSolutionTSP(solutionTSPtemp, solutionVRP.tableauTSP[j]);
		copierSolutionTSP(solutionVRP.tableauTSP[j], solutionVRP.tableauTSP[iMax]);
		copierSolutionTSP(solutionVRP.tableauTSP[iMax], solutionTSPtemp);

		fonctionFree(solutionTSPtemp);
	}
}

int min(int a, int b)
{
	int retour = a;

	if (b < a)
	{
		retour = b;
	}

	return retour;
}

int max(int a, int b)
{
	int retour = a;

	if (b > a)
	{
		retour = b;
	}

	return retour;
}

void goPermuterDansSolutionTSP(solutionTSP_t &solutionTSP, int indiceA, int indiceB)
{
	int temp;
	temp = solutionTSP.etapesTSP[indiceA];
	solutionTSP.etapesTSP[indiceA] = solutionTSP.etapesTSP[indiceB];
	solutionTSP.etapesTSP[indiceB] = temp;
}

void constructionTSPplusProchesVoisinsRandomized(instance_t instance, solutionTSP_t &solutionTSP)
{
	constructionTSPavecTousLesSommetsIndiceCroissant(instance, solutionTSP);
	solutionTSP.nomSolutionTSP = "solutionTSPplusProchesVoisinsRandomized";

	for (int indiceNoeudCourant = 0; indiceNoeudCourant < solutionTSP.nbrNoeudsTSP - 1; indiceNoeudCourant++) // pour tous les noeuds existants
	{
		int tirage = SUR10;
		int nbrTirages = 0;
		while (tirage > PROBAPREMIERCHOIX && nbrTirages < min(NBRPLUSPROCHESVOISINS, solutionTSP.nbrNoeudsTSP - indiceNoeudCourant))
		{
			tirage = (rand() % SUR10) + 1;
			nbrTirages++;
		}

		for (int n = 0; n < nbrTirages; n++)
		{
			int indiceGoPermuter = indiceNoeudPlusProcheVoisinNotVisited(instance, solutionTSP, indiceNoeudCourant + n); // trouver le nieme plus proche voisin du noeud courant

			goPermuterDansSolutionTSP(solutionTSP, indiceGoPermuter, indiceNoeudCourant + 1 + n); // et le permuter avec le nieme suivant du noeud courant
		}

		if (nbrTirages > 1)
		{
			goPermuterDansSolutionTSP(solutionTSP, indiceNoeudCourant + nbrTirages, indiceNoeudCourant + 1); // enfin placer le nieme plus proche voisin à la suite du noeudCourant
		}
	}

	evaluerTSP(instance, solutionTSP, false);
}

void constructionTSPavecTousLesSommetsIndiceCroissant(instance_t instance, solutionTSP_t &solutionTSP)
{
	solutionTSP.nomSolutionTSP = "solutionTSPavecTousLesSommetsIndiceCroissant";
	solutionTSP.nbrNoeudsTSP = 0;

	solutionTSP.etapesTSP[0] = SOMMETDEPOT;
	for (int i = 1; i <= instance.nbrNoeudsInstance; i++)
	{
		solutionTSP.etapesTSP[i] = i;
		solutionTSP.nbrNoeudsTSP++;
	}
	solutionTSP.etapesTSP[solutionTSP.nbrNoeudsTSP + 1] = SOMMETDEPOT;

	evaluerTSP(instance, solutionTSP, false);
}

void constructionTSPcheapestInsertion(instance_t instance, solutionTSP_t &solutionTSP)
{
	// faire listes chaînées pour économiser complexité puisque c'est de l'insertion. mais flemme actuellement

	solutionTSP.nomSolutionTSP = "solutionTSPcheapestInsertion";

	solutionTSP.etapesTSP[0] = SOMMETDEPOT;
	solutionTSP.etapesTSP[1] = (rand() % instance.nbrNoeudsInstance) + 1;
	solutionTSP.etapesTSP[2] = SOMMETDEPOT;

	solutionTSP.nbrNoeudsTSP = 1;

	bool *noeudsRestants;
	noeudsRestants = (bool *)malloc(NBRMAXNOEUDS * sizeof(bool));

	noeudsRestants[SOMMETDEPOT] = false;
	for (int i = 1; i <= instance.nbrNoeudsInstance; i++)
	{
		noeudsRestants[i] = true;
	}
	noeudsRestants[solutionTSP.etapesTSP[1]] = false;
	int nbrNoeudsRestants = instance.nbrNoeudsInstance - 1;

	float **tableauCoutsRelatifs;
	tableauCoutsRelatifs = (float **)malloc(NBRMAXNOEUDS * sizeof(float *));
	for (int i = 0; i < NBRMAXNOEUDS; i++)
	{
		tableauCoutsRelatifs[i] = (float *)malloc(NBRMAXNOEUDS * sizeof(float));
	}

	for (int i = 0; i <= instance.nbrNoeudsInstance; i++)
	{
		if (noeudsRestants[i] == true)
		{
			for (int j = 0; j <= solutionTSP.nbrNoeudsTSP; j++)
			{
				int coutInsertion = instance.distances[solutionTSP.etapesTSP[j]][i] + instance.distances[solutionTSP.etapesTSP[j + 1]][i];
				int coutArrete = instance.distances[solutionTSP.etapesTSP[j]][solutionTSP.etapesTSP[j + 1]];
				float coutRelatifInsertion = (float)coutInsertion / (float)coutArrete;
				tableauCoutsRelatifs[i][j] = coutRelatifInsertion;
			}
		}
	}

	while (nbrNoeudsRestants > 0) // jusqu'à ce que tous les noeuds restants aient été placés
	{
		// recherche insertion la moins coûteuse
		int iMin = 0;
		while (noeudsRestants[iMin] == false)
		{
			iMin++;
		}
		int jMin = 0;
		float coutMin = tableauCoutsRelatifs[iMin][jMin];

		for (int i = iMin; i <= instance.nbrNoeudsInstance; i++)
		{
			if (noeudsRestants[i] == true)
			{
				for (int j = 0; j <= solutionTSP.nbrNoeudsTSP; j++)
				{
					if (tableauCoutsRelatifs[i][j] < coutMin)
					{
						coutMin = tableauCoutsRelatifs[i][j];
						iMin = i;
						jMin = j;
					}
				}
			}
		}

		// application de cette insertion (insertion, décalage dans etapesTSP + MAJ de noeudsRestants[], solutionTSP.nbrNoeudsTSP, coutInsertion[][])
		int noeudRemplaceur = iMin;
		int noeudPourMemoire;
		for (int i = jMin + 1; i <= solutionTSP.nbrNoeudsTSP + 2; i++)
		{
			noeudPourMemoire = solutionTSP.etapesTSP[i];
			solutionTSP.etapesTSP[i] = noeudRemplaceur;
			noeudRemplaceur = noeudPourMemoire;
		}

		solutionTSP.nbrNoeudsTSP++;
		noeudsRestants[iMin] = false;
		nbrNoeudsRestants--;

		for (int i = 0; i <= instance.nbrNoeudsInstance; i++)
		{
			if (noeudsRestants[i] == true)
			{
				float coutRelatifNouveau1 = (float)(instance.distances[solutionTSP.etapesTSP[jMin]][i] + instance.distances[solutionTSP.etapesTSP[jMin + 1]][i]) / (float)instance.distances[solutionTSP.etapesTSP[jMin]][solutionTSP.etapesTSP[jMin + 1]];
				float coutRelatifNouveau2 = (float)(instance.distances[solutionTSP.etapesTSP[jMin + 1]][i] + instance.distances[solutionTSP.etapesTSP[jMin + 2]][i]) / (float)instance.distances[solutionTSP.etapesTSP[jMin + 1]][solutionTSP.etapesTSP[jMin + 2]];

				tableauCoutsRelatifs[i][jMin] = coutRelatifNouveau1;
				float coutRemplaceur = tableauCoutsRelatifs[i][jMin + 1];
				tableauCoutsRelatifs[i][jMin + 1] = coutRelatifNouveau2;
				float coutPourMemoire;
				for (int j = jMin + 2; j <= solutionTSP.nbrNoeudsTSP; j++)
				{
					coutPourMemoire = tableauCoutsRelatifs[i][j];
					tableauCoutsRelatifs[i][j] = coutRemplaceur;
					coutRemplaceur = coutPourMemoire;
				}
			}
		}
		// fin application de cette insertion
	}

	free(noeudsRestants);
	for (int i = 0; i < NBRMAXNOEUDS; i++)
	{
		free(tableauCoutsRelatifs[i]);
	}
	free(tableauCoutsRelatifs);

	evaluerTSP(instance, solutionTSP, false);
}

void SPLIT(instance_t instance, solutionTSP_t &solutionTSP, solutionVRP_t &solutionVRP)
{
	int sizeSolutionTSP = (int)solutionTSP.nomSolutionTSP.size();
	solutionVRP.nomSolutionVRP = "solutionVRP" + solutionTSP.nomSolutionTSP.substr(TAILLEPREFIXESOLUTION, sizeSolutionTSP);

	int indiceCourantNoeudTourGeant = 1;
	int numTourneeCourante = 0;

	while (indiceCourantNoeudTourGeant <= solutionTSP.nbrNoeudsTSP)
	{
		solutionVRP.tableauTSP[numTourneeCourante].nomSolutionTSP = "TSP numéro " + std::to_string(numTourneeCourante) + " de la " + solutionVRP.nomSolutionVRP;

		int capacityDisponibleCourante = instance.capacityVehicle;
		int indiceCourantTourneeCourante = 0;
		solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[indiceCourantTourneeCourante] = SOMMETDEPOT;
		indiceCourantTourneeCourante++;

		while (instance.quantityCollectable[solutionTSP.etapesTSP[indiceCourantNoeudTourGeant]] <= capacityDisponibleCourante && indiceCourantNoeudTourGeant <= solutionTSP.nbrNoeudsTSP)
		{
			solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[indiceCourantTourneeCourante] = solutionTSP.etapesTSP[indiceCourantNoeudTourGeant];
			capacityDisponibleCourante -= instance.quantityCollectable[solutionTSP.etapesTSP[indiceCourantNoeudTourGeant]];
			indiceCourantNoeudTourGeant++;
			indiceCourantTourneeCourante++;
		}

		solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP = indiceCourantTourneeCourante - 1;
		solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[indiceCourantTourneeCourante] = SOMMETDEPOT;

		evaluerTSP(instance, solutionVRP.tableauTSP[numTourneeCourante], false);

		numTourneeCourante++;
	}

	solutionVRP.nbrTourneesVRP = numTourneeCourante;
	evaluerVRP(instance, solutionVRP, false);
}

bool permuterSiRentable(instance_t instance, int indice1, int indice2, solutionTSP_t &solutionTSP)
{
	bool rentableEtPermutated = false;

	solutionTSP_t solutionTSPtemp;
	fonctionMalloc(solutionTSPtemp);

	copierSolutionTSP(solutionTSPtemp, solutionTSP);

	goPermuterDansSolutionTSP(solutionTSPtemp, indice1, indice2);

	evaluerTSP(instance, solutionTSPtemp, false);

	if (solutionTSPtemp.coutTotalTSP < solutionTSP.coutTotalTSP)
	{
		copierSolutionTSP(solutionTSP, solutionTSPtemp);
		rentableEtPermutated = true;
	}

	fonctionFree(solutionTSPtemp);

	return rentableEtPermutated;
}

bool inverserSiRentable(instance_t instance, int indice1, int indice2, solutionTSP_t &solutionTSP)
{
	bool rentableEtInversed = false;

	solutionTSP_t solutionTSPtemp;
	fonctionMalloc(solutionTSPtemp);

	copierSolutionTSP(solutionTSPtemp, solutionTSP);

	int maxIndice = max(indice1, indice2);
	int minIndice = min(indice2, indice1);

	for (int i = minIndice; i <= maxIndice; i++)
	{
		solutionTSPtemp.etapesTSP[minIndice + maxIndice - i] = solutionTSP.etapesTSP[i];
	}

	evaluerTSP(instance, solutionTSPtemp, false);

	if (solutionTSPtemp.coutTotalTSP < solutionTSP.coutTotalTSP)
	{
		copierSolutionTSP(solutionTSP, solutionTSPtemp);
		rentableEtInversed = true;
	}

	fonctionFree(solutionTSPtemp);

	return rentableEtInversed;
}

void RL_2optTSP(instance_t instance, solutionTSP_t &solutionTSP)
{
	if (solutionTSP.nbrNoeudsTSP >= 2)
	{
		bool goRechercher = true;
		while (goRechercher)
		{
			int comboPasDechangement = 0;

			while (comboPasDechangement <= solutionTSP.nbrNoeudsTSP)
			{
				int tirage1 = (rand() % solutionTSP.nbrNoeudsTSP) + 1;
				int tirage2 = (rand() % solutionTSP.nbrNoeudsTSP) + 1;

				if (tirage1 != tirage2)
				{
					bool rentableEtPermutatedOrInversed = permuterSiRentable(instance, tirage1, tirage2, solutionTSP);
					if (!rentableEtPermutatedOrInversed)
					{
						rentableEtPermutatedOrInversed = inverserSiRentable(instance, tirage1, tirage2, solutionTSP);
					}
					comboPasDechangement++;
					if (rentableEtPermutatedOrInversed)
					{
						comboPasDechangement = 0;
					}
				}
			}

			bool pasDeChangement = true;
			int indice1 = 1;
			while (indice1 < solutionTSP.nbrNoeudsTSP && pasDeChangement)
			{
				int indice2 = indice1 + 1;
				while (indice2 <= solutionTSP.nbrNoeudsTSP && pasDeChangement)
				{
					pasDeChangement = !permuterSiRentable(instance, indice1, indice2, solutionTSP);
					if (pasDeChangement)
					{
						pasDeChangement = !inverserSiRentable(instance, indice1, indice2, solutionTSP);
					}
					indice2++;
				}
				indice1++;
			}

			if (pasDeChangement)
			{
				goRechercher = false;
			}
		}
	}
}

void RL_2optVRP(instance_t instance, solutionVRP_t &solutionVRP)
{
	int nbrTotalSommets = 0;
	for (int i = 0; i < solutionVRP.nbrTourneesVRP; i++)
	{
		nbrTotalSommets += solutionVRP.tableauTSP[i].nbrNoeudsTSP;
	}

	if (solutionVRP.nbrTourneesVRP >= 2)
	{
		bool goRechercher = true;

		while (goRechercher)
		{
			int comboPasDechangement = 0;

			while (comboPasDechangement <= nbrTotalSommets)
			{
				comboPasDechangement++;

				int indiceTournee1 = 0;
				int indiceTournee2 = 0;

				while (indiceTournee1 == indiceTournee2)
				{
					indiceTournee1 = rand() % solutionVRP.nbrTourneesVRP;
					indiceTournee2 = rand() % solutionVRP.nbrTourneesVRP;
				}

				int indiceSommetTournee1 = (rand() % solutionVRP.tableauTSP[indiceTournee1].nbrNoeudsTSP) + 1;
				int indiceSommetTournee2 = (rand() % solutionVRP.tableauTSP[indiceTournee2].nbrNoeudsTSP) + 1;

				bool rentableDoncPermutatedEtOptimized = permuterSommetsEtOptimiserTSPsiFaisableEtRentable(instance, solutionVRP.tableauTSP[indiceTournee1], solutionVRP.tableauTSP[indiceTournee2], indiceSommetTournee1, indiceSommetTournee2);

				if (rentableDoncPermutatedEtOptimized)
				{
					comboPasDechangement = 0;
				}
			}

			goRechercher = false;
			int indiceTournee1 = 0;
			while (!goRechercher && indiceTournee1 < solutionVRP.nbrTourneesVRP - 1)
			{
				int indiceSommetTournee1 = 1;
				while (!goRechercher && indiceSommetTournee1 <= solutionVRP.tableauTSP[indiceTournee1].nbrNoeudsTSP)
				{
					int indiceTournee2 = indiceTournee1 + 1;
					while (!goRechercher && indiceTournee2 < solutionVRP.nbrTourneesVRP)
					{
						int indiceSommetTournee2 = 1;
						while (!goRechercher && indiceSommetTournee2 <= solutionVRP.tableauTSP[indiceTournee2].nbrNoeudsTSP)
						{
							goRechercher = permuterSommetsEtOptimiserTSPsiFaisableEtRentable(instance, solutionVRP.tableauTSP[indiceTournee1], solutionVRP.tableauTSP[indiceTournee2], indiceSommetTournee1, indiceSommetTournee2);

							indiceSommetTournee2++;
						}
						indiceTournee2++;
					}
					indiceSommetTournee1++;
				}
				indiceTournee1++;
			}
		}
	}
	evaluerVRP(instance, solutionVRP, false);
}

bool permuterSommetsEtOptimiserTSPsiFaisableEtRentable(instance_t instance, solutionTSP_t &solutionTSP1, solutionTSP_t &solutionTSP2, int indiceSommet1, int indiceSommet2)
{
	bool rentableDoncPermutatedEtOptimized = false;

	int capacityDisponibleTSP1 = instance.capacityVehicle - (solutionTSP1.quantityTotaleTSP - instance.quantityCollectable[solutionTSP1.etapesTSP[indiceSommet1]]) - instance.quantityCollectable[solutionTSP2.etapesTSP[indiceSommet2]];
	int capacityDisponibleTSP2 = instance.capacityVehicle - (solutionTSP2.quantityTotaleTSP - instance.quantityCollectable[solutionTSP2.etapesTSP[indiceSommet2]]) - instance.quantityCollectable[solutionTSP1.etapesTSP[indiceSommet1]];

	if ((capacityDisponibleTSP1 >= 0) && (capacityDisponibleTSP2 >= 0))
	{
		solutionTSP_t solutionTSPtemp1;
		solutionTSP_t solutionTSPtemp2;

		fonctionMalloc(solutionTSPtemp1);
		fonctionMalloc(solutionTSPtemp2);

		copierSolutionTSP(solutionTSPtemp1, solutionTSP1);
		copierSolutionTSP(solutionTSPtemp2, solutionTSP2);

		solutionTSPtemp1.etapesTSP[indiceSommet1] = solutionTSP2.etapesTSP[indiceSommet2];
		solutionTSPtemp2.etapesTSP[indiceSommet2] = solutionTSP1.etapesTSP[indiceSommet1];

		evaluerTSP(instance, solutionTSPtemp1, false);
		evaluerTSP(instance, solutionTSPtemp2, false);

		RL_2optTSP(instance, solutionTSPtemp1);
		RL_2optTSP(instance, solutionTSPtemp2);

		int coutAvant = solutionTSP1.coutTotalTSP + solutionTSP2.coutTotalTSP;
		int coutApres = solutionTSPtemp1.coutTotalTSP + solutionTSPtemp2.coutTotalTSP;

		if (coutApres < coutAvant)
		{
			copierSolutionTSP(solutionTSP1, solutionTSPtemp1);
			copierSolutionTSP(solutionTSP2, solutionTSPtemp2);

			rentableDoncPermutatedEtOptimized = true;
		}

		fonctionFree(solutionTSPtemp1);
		fonctionFree(solutionTSPtemp2);
	}
	return rentableDoncPermutatedEtOptimized;
}

void constructionVRPsommetsLointainsPuisCheapestInsertion(instance_t instance, solutionVRP_t &solutionVRP)
{
	solutionVRP.nomSolutionVRP = "solutionVRPsommetsLointainsPuisCheapestInsertion";
	solutionVRP.nbrTourneesVRP = 0;

	int sommetDistanceMinDepot = 1;
	int distanceMinDepot = instance.distances[SOMMETDEPOT][sommetDistanceMinDepot];
	for (int i = 2; i <= instance.nbrNoeudsInstance; i++)
	{
		if (instance.distances[SOMMETDEPOT][i] < distanceMinDepot)
		{
			sommetDistanceMinDepot = i;
			distanceMinDepot = instance.distances[SOMMETDEPOT][sommetDistanceMinDepot];
		}
	}

	bool sommetsRestants[NBRMAXNOEUDS];
	int nbrNoeudsRestants = instance.nbrNoeudsInstance;
	for (int i = 0; i <= instance.nbrNoeudsInstance; i++)
	{
		sommetsRestants[i] = true;
	}
	sommetsRestants[SOMMETDEPOT] = false;

	bool modeDistanceMaxTrueCapacityMaxFalse = false;
	int numTourneeCourante = 0;
	while (nbrNoeudsRestants > 0)
	{
		// initialisation de cette nouvelle tournée
		solutionVRP.tableauTSP[numTourneeCourante].nomSolutionTSP = "solutionTSPcheapestInsertion";
		solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP = 0;
		solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP = 0;

		int meilleurSommet = 0;

		if (modeDistanceMaxTrueCapacityMaxFalse)
		{
			// Recherche du sommet disponible le plus lointain
			int distanceMaxDepot = instance.distances[SOMMETDEPOT][meilleurSommet];

			for (int indiceSommetInstance = 1; indiceSommetInstance <= instance.nbrNoeudsInstance; indiceSommetInstance++)
			{
				if (sommetsRestants[indiceSommetInstance] == true)
				{
					if (instance.distances[SOMMETDEPOT][indiceSommetInstance] > distanceMaxDepot)
					{
						meilleurSommet = indiceSommetInstance;
						distanceMaxDepot = instance.distances[SOMMETDEPOT][meilleurSommet];
					}
				}
			}
		}
		else
		{
			// Recherche du sommet disponible le plus lourd
			int quantityCollectableMax = instance.quantityCollectable[meilleurSommet];

			for (int indiceSommetInstance = 1; indiceSommetInstance <= instance.nbrNoeudsInstance; indiceSommetInstance++)
			{
				if (sommetsRestants[indiceSommetInstance] == true)
				{
					if (instance.quantityCollectable[indiceSommetInstance] > quantityCollectableMax)
					{
						meilleurSommet = indiceSommetInstance;
						quantityCollectableMax = instance.quantityCollectable[meilleurSommet];
					}
				}
			}
		}

		// insertion du meilleur Sommet trouvé pour le critère donné
		solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[0] = SOMMETDEPOT;
		solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[1] = meilleurSommet;
		solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[2] = SOMMETDEPOT;

		solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP++;
		solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP += instance.quantityCollectable[meilleurSommet];
		sommetsRestants[meilleurSommet] = false;
		nbrNoeudsRestants--;

		int distanceMeilleurSommetAuDepot = instance.distances[SOMMETDEPOT][meilleurSommet];

		// Pour l'instant, le camion fait un aller-retour sans être plein entre le sommet disponible le plus lointain (ou le plus lourd, ça dépend de la variable modeDistanceMaxTrueCapacityMaxFalse) et le dépôt.
		// On complète donc cette tournée avec des sommets qui sont sur le chemin (insertions les moins coûteuses) dans le but de remplir au maximum le camion

		bool continueAdding = true;
		while (continueAdding)
		{
			// initialistaion et remplissage du tableau des coûts relatifs de cette tournée en tenant compte de la capacité disponible du camion
			float tableauCoutsRelatifsPonderated[NBRMAXNOEUDS][NBRMAXNOEUDS];

			bool sommetsEligibles[NBRMAXNOEUDS] = {false};
			int nbrSommetsEligibles = 0;

			for (int i = 0; i <= instance.nbrNoeudsInstance; i++)
			{
				if (sommetsRestants[i] == true) // le sommet est-il disponible ?
				{
					if (instance.quantityCollectable[i] + solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP <= instance.capacityVehicle) // Le sommet ne demande-t-il pas trop de chargement ?
					{
						float sommeDistances = instance.distances[i][SOMMETDEPOT] + instance.distances[i][meilleurSommet];
						// float coefficientEllipse = 1 + (CONSTANTEELLIPSEETCERCLE * distanceMinDepot) / (distanceMeilleurSommetAuDepot * (distanceMeilleurSommetAuDepot - distanceMinDepot)); // formule trouvée et optimisée avec geogebra
						float coefficientEllipse = 1 + (CONSTANTEELLIPSE * distanceMinDepot) / distanceMeilleurSommetAuDepot;

						float rayonCercle = (float)CONSTANTERAYONCERCLE * distanceMeilleurSommetAuDepot;

						if (sommeDistances <= coefficientEllipse * distanceMeilleurSommetAuDepot || (modeDistanceMaxTrueCapacityMaxFalse && ((float)instance.distances[i][meilleurSommet] <= rayonCercle))) // Le sommet est-il situé dans l'ellipse telle que d(sommet, depot) + d(sommet, meilleurSommet) <= COEFFICIENTELLIPSE * d(depot, meilleurSommet) OU, dans le cas où on est en mode "sommet le plus loin disponible", le sommet appartient-il au disque de rayon rayonCercle et de centre le meilleur sommet ?
						{
							sommetsEligibles[i] = true;
							nbrSommetsEligibles++;

							for (int j = 0; j <= solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP; j++)
							{
								int coutInsertion = instance.distances[solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[j]][i] + instance.distances[solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[j + 1]][i];
								int coutArrete = instance.distances[solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[j]][solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[j + 1]];
								float coutRelatifInsertion = (float)coutArrete / (float)coutInsertion;

								tableauCoutsRelatifsPonderated[i][j] = coutRelatifInsertion * instance.quantityCollectable[i] * instance.distances[i][SOMMETDEPOT];
							}
						}
					}
				}
			}

			if (nbrSommetsEligibles == 0)
			{
				continueAdding = false;
			}

			if (continueAdding)
			{
				// recherche de l'insertion la moins coûteuse
				int sommetCheapestInsertion = 0;
				while (sommetsEligibles[sommetCheapestInsertion] == false)
				{
					sommetCheapestInsertion++;
				}

				int indiceArreteIdealePourSommetCheapestInsertion = 0;
				float coutRelatifPonderatedActuelCheapestInsertion = tableauCoutsRelatifsPonderated[sommetCheapestInsertion][indiceArreteIdealePourSommetCheapestInsertion];
				for (int sommetChallenger = sommetCheapestInsertion; sommetChallenger <= instance.nbrNoeudsInstance; sommetChallenger++)
				{
					if (sommetsEligibles[sommetChallenger] == true)
					{
						for (int indiceArreteChallenger = 0; indiceArreteChallenger <= solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP; indiceArreteChallenger++)
						{
							if (tableauCoutsRelatifsPonderated[sommetChallenger][indiceArreteChallenger] > coutRelatifPonderatedActuelCheapestInsertion)
							{
								coutRelatifPonderatedActuelCheapestInsertion = tableauCoutsRelatifsPonderated[sommetChallenger][indiceArreteChallenger];
								sommetCheapestInsertion = sommetChallenger;
								indiceArreteIdealePourSommetCheapestInsertion = indiceArreteChallenger;
							}
						}
					}
				}

				// application de cette insertion -- > insertion et décalage dans les étapes de la tournée
				int sommetRemplaceur = sommetCheapestInsertion;
				int sommetPourMemoire;
				for (int indiceSommetCourant = indiceArreteIdealePourSommetCheapestInsertion + 1; indiceSommetCourant <= solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP + 2; indiceSommetCourant++)
				{
					sommetPourMemoire = solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[indiceSommetCourant];
					solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[indiceSommetCourant] = sommetRemplaceur;
					sommetRemplaceur = sommetPourMemoire;
				}

				solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP++;
				solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP += instance.quantityCollectable[sommetCheapestInsertion];
				sommetsRestants[sommetCheapestInsertion] = false;
				nbrNoeudsRestants--;
			}
		}

		if (!modeDistanceMaxTrueCapacityMaxFalse) // si on est toujours en mode "sommet le plus lourd disponible"
		{
			if ((float)instance.quantityCollectable[meilleurSommet] < PROPORTIONCAPACITYVEHICLE * instance.capacityVehicle) // si le sommet le plus lourd trouvé a occupe moins de la PROPORTIONCAPACITYVEHICLE d'un camion, alors on passe en mode distanceMax jusqu'à la fin
			{
				modeDistanceMaxTrueCapacityMaxFalse = true;
			}
		}

		evaluerTSP(instance, solutionVRP.tableauTSP[numTourneeCourante], false);
		numTourneeCourante++;
	}
	solutionVRP.nbrTourneesVRP = numTourneeCourante;
	evaluerVRP(instance, solutionVRP, false);
}

bool fusionTourneesDansVRP(instance_t instance, solutionVRP_t &solutionVRP)
{
	bool auMoinsUneFusionEffectuee = false;

	// Dans un premier temps, on va trier les tournées par ordre de quantité totale embarquée décroissante
	trierTourneesVRPparQuantityTotaleTSPdecroissante(solutionVRP);

	// Ensuite on va fusionner, si possible, la tournée qui a embarqué le moins avec celle du dessus, puis celle d'encore (si possible) et ainsi de suite..
	int nouveauNbrDeTourneesVRP = solutionVRP.nbrTourneesVRP;
	for (int numTourneeCourante = solutionVRP.nbrTourneesVRP - 1; numTourneeCourante > 0; numTourneeCourante--)
	{
		if (solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP + solutionVRP.tableauTSP[numTourneeCourante - 1].quantityTotaleTSP <= instance.capacityVehicle)
		{
			auMoinsUneFusionEffectuee = true;

			for (int i = 1; i <= solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP + 1; i++)
			{
				solutionVRP.tableauTSP[numTourneeCourante - 1].etapesTSP[solutionVRP.tableauTSP[numTourneeCourante - 1].nbrNoeudsTSP + i] = solutionVRP.tableauTSP[numTourneeCourante].etapesTSP[i];
			}

			solutionVRP.tableauTSP[numTourneeCourante - 1].quantityTotaleTSP += solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP;
			solutionVRP.tableauTSP[numTourneeCourante - 1].nbrNoeudsTSP += solutionVRP.tableauTSP[numTourneeCourante].nbrNoeudsTSP;

			evaluerTSP(instance, solutionVRP.tableauTSP[numTourneeCourante - 1], false);
			solutionVRP.tableauTSP[numTourneeCourante].quantityTotaleTSP = 0; // ainsi, cette tournée va "redescendre" lors du prochain tri
			nouveauNbrDeTourneesVRP--;
		}
	}

	// Pour finir on retrie les tournées de la même façon puis on met à jour le paramètre nbrTorneesVRP
	trierTourneesVRPparQuantityTotaleTSPdecroissante(solutionVRP);
	solutionVRP.nbrTourneesVRP = nouveauNbrDeTourneesVRP;
	evaluerVRP(instance, solutionVRP, false);

	return auMoinsUneFusionEffectuee;
}

void RL_deplacementSommetVRP(instance_t instance, solutionVRP_t &solutionVRP)
{
	int nbrTotalSommets = 0;
	for (int i = 0; i < solutionVRP.nbrTourneesVRP; i++)
	{
		nbrTotalSommets += solutionVRP.tableauTSP[i].nbrNoeudsTSP;
	}

	bool goRechercher = true;
	while (goRechercher)
	{
		int comboPasDechangement = 0;

		while (comboPasDechangement <= nbrTotalSommets)
		{
			comboPasDechangement++;

			int indiceTourneeSource = rand() % solutionVRP.nbrTourneesVRP;
			int indiceTourneeReceveuse = rand() % solutionVRP.nbrTourneesVRP;

			if (indiceTourneeSource < NBRMAXNOEUDS && indiceTourneeSource >= 0 && indiceTourneeReceveuse < NBRMAXNOEUDS && indiceTourneeReceveuse >= 0)
			{
				if ((indiceTourneeSource != indiceTourneeReceveuse && solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP > 0) || (indiceTourneeSource == indiceTourneeReceveuse && solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP >= 2))
				{
					int indiceSommetDeplaced;
					int indiceInsertionSommet;

					if (indiceTourneeSource == indiceTourneeReceveuse)
					{
						indiceSommetDeplaced = (rand() % solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP) + 1;
						indiceInsertionSommet = (rand() % solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP) + 1;

						while (indiceSommetDeplaced == indiceInsertionSommet)
						{
							indiceSommetDeplaced = (rand() % solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP) + 1;
							indiceInsertionSommet = (rand() % solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP) + 1;
						}
					}
					else
					{
						indiceSommetDeplaced = (rand() % solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP) + 1;
						indiceInsertionSommet = (rand() % (solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP + 1)) + 1;
					}

					bool faisableEtRentableDoncDeplacedEtOptimized = false;
					if (indiceTourneeSource == indiceTourneeReceveuse)
					{
						faisableEtRentableDoncDeplacedEtOptimized = deplacerSommetDansTourneeEtOptimiserSiRentable(instance, solutionVRP.tableauTSP[indiceTourneeSource], indiceSommetDeplaced, indiceInsertionSommet);
					}
					else
					{
						faisableEtRentableDoncDeplacedEtOptimized = deplacerSommetDeTournee1DansTournee2EtOptimiserSiFaisableEtRentable(instance, solutionVRP.tableauTSP[indiceTourneeSource], solutionVRP.tableauTSP[indiceTourneeReceveuse], indiceSommetDeplaced, indiceInsertionSommet);
					}

					if (faisableEtRentableDoncDeplacedEtOptimized)
					{
						comboPasDechangement = 0;
					}
				}
			}
		}

		goRechercher = false;
		int indiceTourneeSource = 0;
		while (!goRechercher && indiceTourneeSource < solutionVRP.nbrTourneesVRP)
		{
			int indiceSommetDeplaced = 1;
			while (!goRechercher && indiceSommetDeplaced <= solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP)
			{
				int indiceTourneeReceveuse = 0;
				while (!goRechercher && indiceTourneeReceveuse < solutionVRP.nbrTourneesVRP)
				{
					int indiceInsertionSommet = 1;
					while (!goRechercher && indiceInsertionSommet <= solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP + 1)
					{
						if (indiceTourneeSource == indiceTourneeReceveuse)
						{
							if (indiceSommetDeplaced != indiceInsertionSommet)
							{
								if (indiceInsertionSommet != solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP + 1)
								{
									goRechercher = deplacerSommetDansTourneeEtOptimiserSiRentable(instance, solutionVRP.tableauTSP[indiceTourneeSource], indiceSommetDeplaced, indiceInsertionSommet);
								}
							}
						}
						else
						{
							goRechercher = deplacerSommetDeTournee1DansTournee2EtOptimiserSiFaisableEtRentable(instance, solutionVRP.tableauTSP[indiceTourneeSource], solutionVRP.tableauTSP[indiceTourneeReceveuse], indiceSommetDeplaced, indiceInsertionSommet);
						}
						indiceInsertionSommet++;
					}
					indiceTourneeReceveuse++;
				}
				indiceSommetDeplaced++;
			}
			indiceTourneeSource++;
		}
	}

	evaluerVRP(instance, solutionVRP, false);
}

bool deplacerSommetDeTournee1DansTournee2EtOptimiserSiFaisableEtRentable(instance_t instance, solutionTSP_t &solutionTSP1, solutionTSP_t &solutionTSP2, int indiceSommetTournee, int indiceInsertionSommet)
{
	bool deplacerEtOptimiserSiFaisableRentable = false;

	if (solutionTSP2.quantityTotaleTSP + instance.quantityCollectable[solutionTSP1.etapesTSP[indiceSommetTournee]] <= instance.capacityVehicle)
	{
		solutionTSP_t solutionTSPtemp1;
		solutionTSP_t solutionTSPtemp2;

		fonctionMalloc(solutionTSPtemp1);
		fonctionMalloc(solutionTSPtemp2);

		copierSolutionTSP(solutionTSPtemp1, solutionTSP1);
		copierSolutionTSP(solutionTSPtemp2, solutionTSP2);

		// insérer le sommet au bon indice de solutionTSPtemp2 et décalage à droite des suivants
		for (int i = solutionTSPtemp2.nbrNoeudsTSP + 2; i > indiceInsertionSommet; i--)
		{
			solutionTSPtemp2.etapesTSP[i] = solutionTSPtemp2.etapesTSP[i - 1];
		}
		solutionTSPtemp2.etapesTSP[indiceInsertionSommet] = solutionTSPtemp1.etapesTSP[indiceSommetTournee];
		solutionTSPtemp2.nbrNoeudsTSP++;

		// enlever le sommet de solutionTSPtemp1 et décalage à gauche des sommets suivants
		for (int i = indiceSommetTournee; i <= solutionTSPtemp1.nbrNoeudsTSP; i++)
		{
			solutionTSPtemp1.etapesTSP[i] = solutionTSPtemp1.etapesTSP[i + 1];
		}
		solutionTSPtemp1.nbrNoeudsTSP--;

		evaluerTSP(instance, solutionTSPtemp1, false);
		evaluerTSP(instance, solutionTSPtemp2, false);

		RL_2optTSP(instance, solutionTSPtemp1);
		RL_2optTSP(instance, solutionTSPtemp2);

		int coutAvant = solutionTSP1.coutTotalTSP + solutionTSP2.coutTotalTSP;
		int coutApres = solutionTSPtemp1.coutTotalTSP + solutionTSPtemp2.coutTotalTSP;

		if (coutApres < coutAvant)
		{
			/*
			std::cout << "changement --> cout avant : " << coutAvant << "; cout apres : " << coutApres << "\n";
			if (solutionTSPtemp1.coutTotalTSP > solutionTSP1.coutTotalTSP)
			{
				std::cout << "\n\n\n ALEEEEEEEEEEEEEERTE!!!!!!!!!!!\n\n\n\n" << solutionTSP1.coutTotalTSP << "\n" << solutionTSPtemp1.coutTotalTSP << "\n";

			}
			if (solutionTSPtemp2.coutTotalTSP < solutionTSP2.coutTotalTSP)
			{
				std::cout << "\n\nGIGA Benef\n\n\n";
			}
			*/

			copierSolutionTSP(solutionTSP1, solutionTSPtemp1);
			copierSolutionTSP(solutionTSP2, solutionTSPtemp2);

			deplacerEtOptimiserSiFaisableRentable = true;
		}

		fonctionFree(solutionTSPtemp1);
		fonctionFree(solutionTSPtemp2);
	}

	return deplacerEtOptimiserSiFaisableRentable;
}

bool deplacerSommetDansTourneeEtOptimiserSiRentable(instance_t instance, solutionTSP_t &solutionTSP, int indiceSommetDeplaced, int indiceInsertionSommet)
{
	bool deplacerEtOptimiserSiRentable = false;

	solutionTSP_t solutionTSPtemp;

	fonctionMalloc(solutionTSPtemp);

	copierSolutionTSP(solutionTSPtemp, solutionTSP);

	int sommetDeplaced = solutionTSPtemp.etapesTSP[indiceSommetDeplaced];

	if (indiceSommetDeplaced < indiceInsertionSommet) // décalage à gauche
	{
		for (int i = indiceSommetDeplaced; i < indiceInsertionSommet; i++)
		{
			solutionTSPtemp.etapesTSP[i] = solutionTSPtemp.etapesTSP[i + 1];
		}
	}
	else // normalement pas égal et donc décalage à droite
	{
		for (int i = indiceSommetDeplaced; i > indiceInsertionSommet; i--)
		{
			solutionTSPtemp.etapesTSP[i] = solutionTSPtemp.etapesTSP[i - 1];
		}
	}

	solutionTSPtemp.etapesTSP[indiceInsertionSommet] = sommetDeplaced;

	evaluerTSP(instance, solutionTSPtemp, false);
	RL_2optTSP(instance, solutionTSPtemp);

	if (solutionTSPtemp.coutTotalTSP < solutionTSP.coutTotalTSP)
	{
		copierSolutionTSP(solutionTSP, solutionTSPtemp);

		deplacerEtOptimiserSiRentable = true;
	}

	fonctionFree(solutionTSPtemp);

	return deplacerEtOptimiserSiRentable;
}

void RL_deplacementSommetTSP(instance_t instance, solutionTSP_t &solutionTSP)
{
	bool goRechercher = false;
	if (solutionTSP.nbrNoeudsTSP >= 2)
	{
		goRechercher = true;
	}

	while (goRechercher)
	{
		std::cout << "				Patience...\n";
		int comboPasDechangement = 0;

		while (comboPasDechangement <= solutionTSP.nbrNoeudsTSP)
		{
			comboPasDechangement++;

			int indiceSommetDeplaced = 0;
			int indiceInsertionSommet = 0;

			while (indiceSommetDeplaced == indiceInsertionSommet) // pas infini car condition de départ --> minimum 2 sommets
			{
				indiceSommetDeplaced = (rand() % solutionTSP.nbrNoeudsTSP) + 1;
				indiceInsertionSommet = (rand() % solutionTSP.nbrNoeudsTSP) + 1;
			}

			bool rentableDoncDeplacedEtOptimized = deplacerSommetDansTourneeEtOptimiserSiRentable(instance, solutionTSP, indiceSommetDeplaced, indiceInsertionSommet);

			if (rentableDoncDeplacedEtOptimized)
			{
				comboPasDechangement = 0;
			}
		}

		goRechercher = false;
		int indiceSommetDeplaced = 1;
		while (!goRechercher && indiceSommetDeplaced <= solutionTSP.nbrNoeudsTSP)
		{
			int coutDetourParCeSommet = instance.distances[solutionTSP.etapesTSP[indiceSommetDeplaced - 1]][solutionTSP.etapesTSP[indiceSommetDeplaced]] + instance.distances[solutionTSP.etapesTSP[indiceSommetDeplaced]][solutionTSP.etapesTSP[indiceSommetDeplaced + 1]];
			int coutSansDetour = instance.distances[solutionTSP.etapesTSP[indiceSommetDeplaced - 1]][solutionTSP.etapesTSP[indiceSommetDeplaced + 1]];

			if ((float)coutDetourParCeSommet > 1.4 * coutSansDetour)
			{
				int indiceInsertionSommet = 1;
				while (!goRechercher && indiceInsertionSommet <= solutionTSP.nbrNoeudsTSP)
				{
					if (indiceSommetDeplaced != indiceInsertionSommet)
					{
						goRechercher = deplacerSommetDansTourneeEtOptimiserSiRentable(instance, solutionTSP, indiceSommetDeplaced, indiceInsertionSommet);
					}
					indiceInsertionSommet++;
				}
			}
			indiceSommetDeplaced++;
		}
	}
}

void productionDeToursGeantsPourSPLIT(contenantTableauSolutionsTSP_t &contenantTableauSolutionsTSPpourSPLIT, instance_t instance)
{
	std::cout << "	Run de la contruction initiales des tours géants (solutionTSP)\n";

	contenantTableauSolutionsTSPpourSPLIT.nomTableauSolutionsTSP = "contenantTableauSolutionsTSPpourSPLIT";
	contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP = 0;

	// Construction initiale des tours géants avec plusieurs stratégies
	for (int i = 0; i < NBRALGOSCONSTRUCTION_TSP; i++)
	{
		switch (i)
		{
		case 0:
			constructionTSPplusProcheVoisins(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		case 1:
			constructionTSPplusProchesVoisinsRandomized(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		case 2:
			constructionTSPcheapestInsertion(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		case 3:
			constructionTSPavecTousLesSommetsIndiceCroissant(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		case 4:
			constructionTSPplusLointainVoisin(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		case 5:
			constructionTSPrandom(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			break;
		}
	}
	contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP += NBRALGOSCONSTRUCTION_TSP;


	std::cout << "	Run de l'optimisation de ces tours géants : \n		Recherche locale par permutation ou inversion\n";


	//contenantTableauSolutionsTSPpourSPLIT.nomTableauSolutionsTSP += "_optiTSP";
	for (int i = 0; i < contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP; i++)
	{
		RL_2optTSP(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
	}


	if(GROSSEOPTI_TSP_INITIALE) // il ne faut pas hésiter à exécuter avec GROSSEOPTI_TSP_INITIALE = false si le temps d'exécution est trop long (grande instance), car cette fonctionnalité est très gourmande, pas très utile et pas paramètrable.
	{
		std::cout << "		Recherche locale par déplacement :\n";
		//contenantTableauSolutionsTSPpourSPLIT.nomTableauSolutionsTSP += "_RLdeplacementSommetsTSP";
		for (int i = 0; i < contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP; i++)
		{
			std::cout << "			Tour géant d'indice " << i << ", plus que " << contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP - 1 - i << " (procédure assez longue)\n";
			RL_deplacementSommetTSP(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
		}
	}else{
		std::cout << "		Pas de recherche locale par déplacement --> économie de temps d'exécution.\n";
	}
	
	
	enleverDoublonsEtAjouterInverses(contenantTableauSolutionsTSPpourSPLIT);
}

void enleverDoublonsEtAjouterInverses(contenantTableauSolutionsTSP_t &contenantTableauSolutionsTSPpourSPLIT)
{
	contenantTableauSolutionsTSP_t contenantTableauSolutionsTSPtemp;
	fonctionMalloc(contenantTableauSolutionsTSPtemp, TAILLEMAXCONTENANTSOLUTION_TSP);
	contenantTableauSolutionsTSPtemp.nomTableauSolutionsTSP = contenantTableauSolutionsTSPpourSPLIT.nomTableauSolutionsTSP;
	contenantTableauSolutionsTSPtemp.nbrSolutionsTSP = 0;

	// Checker si il y a des doublons (solutions identiques ou inversées) et si oui les virer
	for (int i = 0; i < contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP; i++)
	{
		int j = 0;
		bool equalOrInversed = false;
		while (i != 0 && !equalOrInversed && j < contenantTableauSolutionsTSPtemp.nbrSolutionsTSP)
		{
			equalOrInversed = solutionTSP1_equalOrInversed_SolutionTSP2(contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i], contenantTableauSolutionsTSPtemp.tableauSolutionsTSP[j]);
			j++;
		}

		if (!equalOrInversed)
		{
			copierSolutionTSP(contenantTableauSolutionsTSPtemp.tableauSolutionsTSP[j], contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i]);
			contenantTableauSolutionsTSPtemp.nbrSolutionsTSP++;
		}
	}


	// Pour chaque solution retenue, on ajoute également son inverse pour tout tenter avec le SPLIT
	for (int i = 0; i < contenantTableauSolutionsTSPtemp.nbrSolutionsTSP; i++)
	{
		solutionTSP_t solutionTSPinversee;
		fonctionMalloc(solutionTSPinversee);
		copierSolutionTSP(solutionTSPinversee, contenantTableauSolutionsTSPtemp.tableauSolutionsTSP[i]);

		solutionTSPinversee.nomSolutionTSP += "_inverse";
		for (int indice = 0; indice <= solutionTSPinversee.nbrNoeudsTSP + 1; indice++)
		{
			solutionTSPinversee.etapesTSP[indice] = contenantTableauSolutionsTSPtemp.tableauSolutionsTSP[i].etapesTSP[solutionTSPinversee.nbrNoeudsTSP + 1 - indice];
		}
		copierSolutionTSP(contenantTableauSolutionsTSPtemp.tableauSolutionsTSP[contenantTableauSolutionsTSPtemp.nbrSolutionsTSP + i], solutionTSPinversee);

		fonctionFree(solutionTSPinversee);
	}
	contenantTableauSolutionsTSPtemp.nbrSolutionsTSP *= 2;

	copierContenantTableauSolutionTSP(contenantTableauSolutionsTSPpourSPLIT, contenantTableauSolutionsTSPtemp);
	fonctionFree(contenantTableauSolutionsTSPtemp, TAILLEMAXCONTENANTSOLUTION_TSP);
}

void productionDeVRPpour1erCycleOpti(contenantTableauSolutionsTSP_t &contenantTableauSolutionsTSPpourSPLIT, contenantTableauSolutionsVRP_t &contenantTableauSolutionsVRPpourOPTI, instance_t instance)
{
	contenantTableauSolutionsVRPpourOPTI.nomTableauSolutionsVRP = "contenantTableauSolutionsVRPpour1erCycle";
	contenantTableauSolutionsVRPpourOPTI.nbrSolutionsVRP = 0;

	for (int i = 0; i < contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP; i++)
	{
		SPLIT(instance, contenantTableauSolutionsTSPpourSPLIT.tableauSolutionsTSP[i], contenantTableauSolutionsVRPpourOPTI.tableauSolutionsVRP[i]);
	}

	for (int i = 0; i < NBRALGOSCONSTRUCTION_VRP; i++)
	{
		switch (i)
		{
		case 0:
			constructionVRPsommetsLointainsPuisCheapestInsertion(instance, contenantTableauSolutionsVRPpourOPTI.tableauSolutionsVRP[contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP + i]);
			break;
		}
	}

	contenantTableauSolutionsVRPpourOPTI.nbrSolutionsVRP += contenantTableauSolutionsTSPpourSPLIT.nbrSolutionsTSP + NBRALGOSCONSTRUCTION_VRP;
}

bool solutionTSP1_equalOrInversed_SolutionTSP2(solutionTSP_t solutionTSP1, solutionTSP_t solutionTSP2)
{
	bool equalOrInversed = false;

	if (solutionTSP1.nbrNoeudsTSP == solutionTSP2.nbrNoeudsTSP)
	{
		if (solutionTSP1.coutTotalTSP == solutionTSP2.coutTotalTSP)
		{
			bool areEqual = true;
			bool areInversed = true;
			int indiceCourant = 0;

			while ((areEqual || areInversed) && indiceCourant <= solutionTSP1.nbrNoeudsTSP + 1)
			{
				if (areEqual)
				{
					if (solutionTSP1.etapesTSP[indiceCourant] != solutionTSP2.etapesTSP[indiceCourant])
					{
						areEqual = false;
					}
				}
				if (areInversed)
				{
					if (solutionTSP1.etapesTSP[indiceCourant] != solutionTSP2.etapesTSP[solutionTSP1.nbrNoeudsTSP + 1 - indiceCourant])
					{
						areInversed = false;
					}
				}
				indiceCourant++;
			}

			if ((areEqual && !areInversed) || (!areEqual && areInversed))
			{
				equalOrInversed = true;
			}
		}
	}

	return equalOrInversed;
}

void procedure1erCycleOptimisation(instance_t instance, contenantTableauSolutionsVRP_t contenantTableauSolutionsVRPpourOpti, contenantTableauSolutionsVRP_t &meilleuresSolutionsVRPtrouvees)
{
	contenantTableauSolutionsVRP_t contenantTableauSolutionsVRPtemp;
	fonctionMalloc(contenantTableauSolutionsVRPtemp, contenantTableauSolutionsVRPpourOpti.nbrSolutionsVRP);
	copierContenantTableauSolutionVRP(contenantTableauSolutionsVRPtemp, contenantTableauSolutionsVRPpourOpti);

	contenantTableauSolutionsVRPtemp.nomTableauSolutionsVRP += "_fusionTournees";
	for (int i = 0; i < contenantTableauSolutionsVRPtemp.nbrSolutionsVRP; i++)
	{
		fusionTourneesDansVRP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i]);
	}

	bool auMoinsUneFusionEffectuee = true;

	while (auMoinsUneFusionEffectuee)
	{
		auMoinsUneFusionEffectuee = false;

		contenantTableauSolutionsVRPtemp.nomTableauSolutionsVRP += "_optiVRP";
		for (int i = 0; i < contenantTableauSolutionsVRPtemp.nbrSolutionsVRP; i++)
		{
			for (int j = 0; j < contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i].nbrTourneesVRP; j++)
			{
				RL_2optTSP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i].tableauTSP[j]);
			}
			evaluerVRP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i], false);
		}

		contenantTableauSolutionsVRPtemp.nomTableauSolutionsVRP += "_optiInterVRP";
		for (int i = 0; i < contenantTableauSolutionsVRPtemp.nbrSolutionsVRP; i++)
		{
			RL_2optVRP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i]);
		}

		contenantTableauSolutionsVRPtemp.nomTableauSolutionsVRP += "_optiDeplacementVRP";
		for (int i = 0; i < contenantTableauSolutionsVRPtemp.nbrSolutionsVRP; i++)
		{
			RL_deplacementSommetVRP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i]);
		}


		contenantTableauSolutionsVRPtemp.nomTableauSolutionsVRP += "_fusionTournees";
		for (int i = 0; i < contenantTableauSolutionsVRPtemp.nbrSolutionsVRP; i++)
		{
			if (fusionTourneesDansVRP(instance, contenantTableauSolutionsVRPtemp.tableauSolutionsVRP[i]))
			{
				auMoinsUneFusionEffectuee = true;
			}
		}
	}

	copierContenantTableauSolutionVRP(meilleuresSolutionsVRPtrouvees, contenantTableauSolutionsVRPtemp);
	fonctionFree(contenantTableauSolutionsVRPtemp, contenantTableauSolutionsVRPpourOpti.nbrSolutionsVRP);
}

void procedureOptiCompleteSolutionVRP(instance_t instance, solutionVRP_t &solutionVRP)
{
	fusionTourneesDansVRP(instance, solutionVRP);

	bool auMoinsUneFusionEffectuee = true;

	while (auMoinsUneFusionEffectuee)
	{
		for (int i = 0; i < solutionVRP.nbrTourneesVRP; i++)
		{
			RL_2optTSP(instance, solutionVRP.tableauTSP[i]);
		}
		evaluerVRP(instance, solutionVRP, false);

		RL_2optVRP(instance, solutionVRP);

		RL_deplacementSommetVRP(instance, solutionVRP);

		auMoinsUneFusionEffectuee = fusionTourneesDansVRP(instance, solutionVRP);
	}
}


void procedure2emeCycleOptimisation(instance_t instance, contenantTableauSolutionsVRP_t contenantTableauSolutionsVRPpourOptiAvecRegression, contenantTableauSolutionsVRP_t &meilleuresSolutionsVRPtrouvees)
{
	for (int i = 0; i < contenantTableauSolutionsVRPpourOptiAvecRegression.nbrSolutionsVRP; i++)
	{
		std::cout << "\n	Run 2ème cycle optimisation sur la solution VRP d'indice " << i << " (ça va jusqu'à l'indice " << contenantTableauSolutionsVRPpourOptiAvecRegression.nbrSolutionsVRP - 1 << ") de la sélection des " << NBRSOLUTIONSVRPPOURDEUXIEMECYCLE << " meilleures solutions VRP trouvées par le 1er cycle d'optimisation\n	";
		afficherSolutionVRP(contenantTableauSolutionsVRPpourOptiAvecRegression.tableauSolutionsVRP[i]);

		solutionVRP_t solutionVRPcourante;
		solutionVRP_t solutionVRPmeilleureTrouvee;
		fonctionMalloc(solutionVRPcourante);
		fonctionMalloc(solutionVRPmeilleureTrouvee);

		copierSolutionVRP(solutionVRPcourante, contenantTableauSolutionsVRPpourOptiAvecRegression.tableauSolutionsVRP[i]);
		copierSolutionVRP(solutionVRPmeilleureTrouvee, solutionVRPcourante);

		for (int j = 0; j < NBRDERUNSREGRESSIONRECURSIVE; j++)
		{
			std::cout << "\n		Run ayant pour objectif de trouver la " << j + 1 << " ième régression améliorante, on veut minimum " << NBRDERUNSREGRESSIONRECURSIVE << " régressions améliorantes\n";

			bool continuer = true;
			int profondeurActuelle = 0;

			effectuerRegressionRecursive(instance, solutionVRPcourante, solutionVRPmeilleureTrouvee, profondeurActuelle, continuer);
			copierSolutionVRP(solutionVRPcourante, solutionVRPmeilleureTrouvee);
		}

		copierSolutionVRP(meilleuresSolutionsVRPtrouvees.tableauSolutionsVRP[i], solutionVRPmeilleureTrouvee);
		meilleuresSolutionsVRPtrouvees.nbrSolutionsVRP++;

		fonctionFree(solutionVRPcourante);
		fonctionFree(solutionVRPmeilleureTrouvee);
	}
}

void effectuerRegression(instance_t instance, solutionVRP_t solutionVRP/*, float deltaAccepted*/)
{
	if (solutionVRP.nbrTourneesVRP >= 1)
	{
		if (solutionVRP.nbrTourneesVRP > 1 || (solutionVRP.nbrTourneesVRP == 1 && solutionVRP.tableauTSP[0].nbrNoeudsTSP >= 2))
		{
			bool regressionEffectuee = false;
			while (!regressionEffectuee)
			{
				solutionVRP_t solutionVRPtemp;
				fonctionMalloc(solutionVRPtemp);
				copierSolutionVRP(solutionVRPtemp, solutionVRP);

				int tirageAlgo = rand() % NBRALGOSPERTURBATIONS;
				if (tirageAlgo == PERMUTER)
				{
					int tirageTournee1;
					int tirageTournee2;
					int tirageIndiceSommetTournee1;
					int tirageIndiceSommetTournee2;

					trouverPermutationFaisableDansSolutionVRP(instance, solutionVRP, tirageTournee1, tirageTournee2, tirageIndiceSommetTournee1, tirageIndiceSommetTournee2);

					if (tirageTournee1 == tirageTournee2)
					{
						goPermuterDansSolutionTSP(solutionVRPtemp.tableauTSP[tirageTournee1], tirageIndiceSommetTournee1, tirageIndiceSommetTournee2);
						evaluerTSP(instance, solutionVRPtemp.tableauTSP[tirageTournee1], false);
					}
					else
					{
						int sommetTemp = solutionVRPtemp.tableauTSP[tirageTournee1].etapesTSP[tirageIndiceSommetTournee1];
						solutionVRPtemp.tableauTSP[tirageTournee1].etapesTSP[tirageIndiceSommetTournee1] = solutionVRPtemp.tableauTSP[tirageTournee2].etapesTSP[tirageIndiceSommetTournee2];
						solutionVRPtemp.tableauTSP[tirageTournee2].etapesTSP[tirageIndiceSommetTournee2] = sommetTemp;
						evaluerTSP(instance, solutionVRPtemp.tableauTSP[tirageTournee1], false);
						evaluerTSP(instance, solutionVRPtemp.tableauTSP[tirageTournee2], false);
					}

					evaluerVRP(instance, solutionVRPtemp, false);

					//float deltaPerturbation = solutionVRPtemp.coutTotalVRP - solutionVRP.coutTotalVRP;
					//if (deltaPerturbation <= deltaAccepted)
					//{
						copierSolutionVRP(solutionVRP, solutionVRPtemp);
						regressionEffectuee = true;
					//}
				}
				else if (tirageAlgo == INVERSER)
				{
					int tirageTournee = rand() % solutionVRP.nbrTourneesVRP;

					if (solutionVRP.tableauTSP[tirageTournee].nbrNoeudsTSP >= 2)
					{
						int tirageIndiceSommet1 = 0;
						int tirageIndiceSommet2 = 0;

						while (tirageIndiceSommet1 == tirageIndiceSommet2)
						{
							tirageIndiceSommet1 = (rand() % solutionVRP.tableauTSP[tirageTournee].nbrNoeudsTSP) + 1;
							tirageIndiceSommet2 = (rand() % solutionVRP.tableauTSP[tirageTournee].nbrNoeudsTSP) + 1;
						}

						int maxIndice = max(tirageIndiceSommet1, tirageIndiceSommet2);
						int minIndice = min(tirageIndiceSommet1, tirageIndiceSommet2);

						for (int i = minIndice; i <= maxIndice; i++)
						{
							solutionVRPtemp.tableauTSP[tirageTournee].etapesTSP[minIndice + maxIndice - i] = solutionVRP.tableauTSP[tirageTournee].etapesTSP[i];
						}

						evaluerTSP(instance, solutionVRPtemp.tableauTSP[tirageTournee], false);

						evaluerVRP(instance, solutionVRPtemp, false);

						//float deltaPerturbation = solutionVRPtemp.coutTotalVRP - solutionVRP.coutTotalVRP;
						//if (deltaPerturbation <= deltaAccepted)
						//{
							copierSolutionVRP(solutionVRP, solutionVRPtemp);
							regressionEffectuee = true;
						//}
					}
				}
				else if (tirageAlgo == DEPLACER)
				{
					int indiceTourneeSource = rand() % solutionVRP.nbrTourneesVRP;
					int indiceTourneeReceveuse = rand() % solutionVRP.nbrTourneesVRP;

					if (indiceTourneeSource < NBRMAXNOEUDS && indiceTourneeSource >= 0 && indiceTourneeReceveuse < NBRMAXNOEUDS && indiceTourneeReceveuse >= 0)
					{
						if ((indiceTourneeSource != indiceTourneeReceveuse && solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP > 0) || (indiceTourneeSource == indiceTourneeReceveuse && solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP >= 2))
						{
							int indiceSommetDeplaced;
							int indiceInsertionSommet;

							if (indiceTourneeSource == indiceTourneeReceveuse)
							{
								indiceSommetDeplaced = 0;
								indiceInsertionSommet = 0;

								while (indiceSommetDeplaced == indiceInsertionSommet)
								{
									indiceSommetDeplaced = (rand() % solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP) + 1;
									indiceInsertionSommet = (rand() % solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP) + 1;
								}

								int sommetDeplaced = solutionVRP.tableauTSP[indiceTourneeSource].etapesTSP[indiceSommetDeplaced];

								if (indiceSommetDeplaced < indiceInsertionSommet) // décalage à gauche
								{
									for (int i = indiceSommetDeplaced; i < indiceInsertionSommet; i++)
									{
										solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i] = solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i + 1];
									}
								}
								else // normalement pas égal et donc décalage à droite
								{
									for (int i = indiceSommetDeplaced; i > indiceInsertionSommet; i--)
									{
										solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i] = solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i - 1];
									}
								}

								solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[indiceInsertionSommet] = sommetDeplaced;

								evaluerTSP(instance, solutionVRPtemp.tableauTSP[indiceTourneeSource], false);
							}
							else
							{
								indiceSommetDeplaced = (rand() % solutionVRP.tableauTSP[indiceTourneeSource].nbrNoeudsTSP) + 1;
								indiceInsertionSommet = (rand() % (solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP + 1)) + 1;

								if (solutionVRP.tableauTSP[indiceTourneeReceveuse].quantityTotaleTSP + instance.quantityCollectable[solutionVRP.tableauTSP[indiceTourneeSource].etapesTSP[indiceSommetDeplaced]] <= instance.capacityVehicle)
								{
									for (int i = solutionVRP.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP + 2; i > indiceInsertionSommet; i--)
									{
										solutionVRPtemp.tableauTSP[indiceTourneeReceveuse].etapesTSP[i] = solutionVRPtemp.tableauTSP[indiceTourneeReceveuse].etapesTSP[i - 1];
									}
									solutionVRPtemp.tableauTSP[indiceTourneeReceveuse].etapesTSP[indiceInsertionSommet] = solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[indiceSommetDeplaced];
									solutionVRPtemp.tableauTSP[indiceTourneeReceveuse].nbrNoeudsTSP++;

									// enlever le sommet de solutionTSPtemp1 et décalage à gauche des sommets suivants
									for (int i = indiceSommetDeplaced; i <= solutionVRPtemp.tableauTSP[indiceTourneeSource].nbrNoeudsTSP; i++)
									{
										solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i] = solutionVRPtemp.tableauTSP[indiceTourneeSource].etapesTSP[i + 1];
									}
									solutionVRPtemp.tableauTSP[indiceTourneeSource].nbrNoeudsTSP--;

									evaluerTSP(instance, solutionVRPtemp.tableauTSP[indiceTourneeSource], false);
									evaluerTSP(instance, solutionVRPtemp.tableauTSP[indiceTourneeReceveuse], false);
								}
							}

							evaluerVRP(instance, solutionVRPtemp, false);

							//float deltaPerturbation = solutionVRPtemp.coutTotalVRP - solutionVRP.coutTotalVRP;
							//if (deltaPerturbation <= deltaAccepted)
							//{
								copierSolutionVRP(solutionVRP, solutionVRPtemp);
								regressionEffectuee = true;
							//}
						}
					}
				}

				fonctionFree(solutionVRPtemp);
			}
		}
	}
}

void trouverPermutationFaisableDansSolutionVRP(instance_t instance, solutionVRP_t &solutionVRP, int &tirageTournee1, int &tirageTournee2, int &tirageIndiceSommetTournee1, int &tirageIndiceSommetTournee2)
{
	bool capacityCompatible = false;

	while (!capacityCompatible)
	{
		bool continuerTiragesTournees = true;
		while (continuerTiragesTournees)
		{
			continuerTiragesTournees = false;

			tirageTournee1 = rand() % solutionVRP.nbrTourneesVRP;
			tirageTournee2 = rand() % solutionVRP.nbrTourneesVRP;

			if (solutionVRP.tableauTSP[tirageTournee1].nbrNoeudsTSP < 1 || solutionVRP.tableauTSP[tirageTournee2].nbrNoeudsTSP < 1 || (tirageTournee1 == tirageTournee2 && solutionVRP.tableauTSP[tirageTournee1].nbrNoeudsTSP < 2))
			{
				continuerTiragesTournees = true;
			}
		}

		bool continuerTirageSommetDansTournees = true;
		while (continuerTirageSommetDansTournees)
		{
			continuerTirageSommetDansTournees = false;

			tirageIndiceSommetTournee1 = (rand() % solutionVRP.tableauTSP[tirageTournee1].nbrNoeudsTSP) + 1;
			tirageIndiceSommetTournee2 = (rand() % solutionVRP.tableauTSP[tirageTournee2].nbrNoeudsTSP) + 1;

			if (tirageTournee1 == tirageTournee2 && tirageIndiceSommetTournee1 == tirageIndiceSommetTournee2)
			{
				continuerTirageSommetDansTournees = true;
			}
		}

		int capacityDisponibleTournee1 = instance.capacityVehicle - (solutionVRP.tableauTSP[tirageTournee1].quantityTotaleTSP - instance.quantityCollectable[solutionVRP.tableauTSP[tirageTournee1].etapesTSP[tirageIndiceSommetTournee1]]) - instance.quantityCollectable[solutionVRP.tableauTSP[tirageTournee2].etapesTSP[tirageIndiceSommetTournee2]];
		int capacityDisponibleTournee2 = instance.capacityVehicle - (solutionVRP.tableauTSP[tirageTournee2].quantityTotaleTSP - instance.quantityCollectable[solutionVRP.tableauTSP[tirageTournee2].etapesTSP[tirageIndiceSommetTournee2]]) - instance.quantityCollectable[solutionVRP.tableauTSP[tirageTournee1].etapesTSP[tirageIndiceSommetTournee1]];

		if ((capacityDisponibleTournee1 >= 0) && (capacityDisponibleTournee2 >= 0))
		{
			capacityCompatible = true;
		}
	}
}

void effectuerRegressionRecursive(instance_t instance, solutionVRP_t solutionVRPcourante, solutionVRP_t &solutionVRPmeilleureTrouvee, int profondeurActuelle, bool &continuer)
{
	solutionVRP_t solutionVRPtemp;
	fonctionMalloc(solutionVRPtemp);
	copierSolutionVRP(solutionVRPtemp, solutionVRPcourante);

	effectuerRegression(instance, solutionVRPtemp);
	procedureOptiCompleteSolutionVRP(instance, solutionVRPtemp);

	if (solutionVRPtemp.coutTotalVRP < solutionVRPmeilleureTrouvee.coutTotalVRP)
	{
		copierSolutionVRP(solutionVRPmeilleureTrouvee, solutionVRPtemp);
		std::cout << "		";
		afficherSolutionVRP(solutionVRPtemp);
		continuer = false;
	}

	if (continuer && profondeurActuelle < PROFONDEURREGRESSIONRECURSIVE)
	{
		for (int i = 0; i < NBRBRANCHESREGRESSIONRECURSIVE; i++)
		{
			effectuerRegressionRecursive(instance, solutionVRPtemp, solutionVRPmeilleureTrouvee, profondeurActuelle + 1, continuer);
		}
	}
	else
	{
		long int grandNombreAleatoire = 0;
		for(int i = 0; i < 50; i++)// ainsi, grandChiffreAleatoire est distribué uniformément sur [0, 2^50 - 1], je pense que c'est suffisant haha
		{
			grandNombreAleatoire *= 2; 
			grandNombreAleatoire += rand() % 2;
		}
		int tirage = grandNombreAleatoire % (int)PROBAPRINTPATIENCE; // normalement il devrait y avoir ~4 "Patience" pour une run avec NBRBRANCHESREGRESSIONRECURSIVE = 10
		if(tirage == 0)
		{
			std::cout << "		Patience ...\n";
		}
	}
	

	fonctionFree(solutionVRPtemp);
}